//
//  LKGeometry.h
//  Vectors
//
//  Created by Károly Lőrentey on 2010-10-17.
//  Copyright 2010 Károly Lőrentey. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <CoreGraphics/CoreGraphics.h>
#import <UIKit/UIKit.h>

enum CGRectCorner {
    CGRectTopLeft,
    CGRectTopCenter,
    CGRectTopRight,
    CGRectCenterLeft,
    CGRectCenterCenter,
    CGRectCenterRight,
    CGRectBottomLeft,
    CGRectBottomCenter,
    CGRectBottomRight
};
typedef enum CGRectCorner CGRectCorner;

typedef struct LKLine {
    CGPoint pt;
    CGSize vk;
} LKLine;

typedef struct LKSegment {
    CGPoint ptStart;
    CGPoint ptEnd;
} LKSegment;


extern CGAffineTransform CGAffineTransformMakeScaleAround(float xScale, float yScale, CGPoint center);

#pragma mark -
#pragma mark Rectangles

static inline CGPoint CGPointAdd(CGPoint pt, CGSize vk)
{
    return CGPointMake(pt.x + vk.width, pt.y + vk.height);
}

static inline CGPoint CGPointSubtract(CGPoint pt, CGSize vk)
{
    return CGPointMake(pt.x - vk.width, pt.y - vk.height);
}

extern CGFloat CGPointDistance(CGPoint pt1, CGPoint pt2);
extern CGPoint CGPointScale(CGPoint pt, float scale);
extern CGPoint CGPointInterpolate(float t, CGPoint pt1, CGPoint pt2);

static inline CGSize CGSizeFromCGPoint(CGPoint pt)
{
    return CGSizeMake(pt.x, pt.y);
}

static inline CGSize CGSizeBetween(CGPoint pt1, CGPoint pt2)
{
    return CGSizeMake(pt2.x - pt1.x, pt2.y - pt1.y);
}

static inline float CGSizeDotProduct(CGSize vk1, CGSize vk2)
{
    return vk1.width * vk2.width + vk1.height * vk2.height;
}

extern float CGSizeLength(CGSize size);
extern CGSize CGSizeScale(CGSize size, float scale);

extern float CGSizeAngleBetween(CGSize vk1, CGSize vk2); // in radians

static inline CGPoint
CGRectGetMidPoint(CGRect rect)
{
    return CGPointMake(rect.origin.x + rect.size.width / 2.0f,
                       rect.origin.y + rect.size.height / 2.0f);
}

static inline CGRect
CGRectAround(CGPoint center, float radiusX, float radiusY)
{
    return CGRectMake(center.x - radiusX, center.y - radiusY, 2 * radiusX, 2 * radiusY);
}

extern CGRect CGRectRoundToInteger(CGRect rect);

extern CGRect CGRectScaleAroundOrigin(CGRect rect, float scaleX, float scaleY);
extern CGRect CGRectScaleAroundMiddle(CGRect rect, float scaleX, float scaleY);

// Extend RECT if necessary such that its aspect matches that of SIZE.
// The original rect is centered in the result.
extern CGRect CGRectExtendToFitAspect(CGRect rect, CGSize size);

// Move rect to fit inside boundaries of bounds.
extern CGRect CGRectMoveToFitInRect(CGRect rect, CGRect bounds);

extern void CGPathAddRoundRect(CGMutablePathRef path, CGRect rect, float radius);
extern void CGContextAddRoundRect(CGContextRef context, CGRect rect, float radius);

extern CGPoint CGRectGetCorner(CGRect rect, CGRectCorner corner);
extern void CGRectSetCorner(CGRect *rect, CGRectCorner corner, CGPoint point);

// YES if rcInner is inside rcOuter, not touching any of its edges.
extern BOOL CGRectInsideRect(CGRect rcInner, CGRect rcOuter);

extern NSString *CGRectString(CGRect rect);

#pragma mark -
#pragma mark Insets

inline static UIEdgeInsets
UIEdgeInsetsScale(UIEdgeInsets insets, float scaleX, float scaleY)
{
    return UIEdgeInsetsMake(scaleY * insets.top, scaleX * insets.left, scaleY * insets.bottom, scaleX * insets.right);
}

#pragma mark -
#pragma mark Lines

extern LKLine LKLineBetween(CGPoint p1, CGPoint p2);

inline static CGPoint
CGPointOnLine(LKLine ln, float t)
{
    return CGPointAdd(ln.pt, CGSizeScale(ln.vk, t));
}

extern CGPoint CGPointClosestOnLine(LKLine ln, CGPoint pt);
extern float LKLineParameterClosestOnLine(LKLine ln, CGPoint pt);

extern BOOL CGPointIsToTheRight(LKLine ln, CGPoint pt);
extern CGPoint CGPointIntersection(LKLine ln1, LKLine ln2);

extern BOOL LKLineIntersectsRect(LKLine ln, CGRect rc);
extern LKSegment LKSegmentIntersectionLineRect(LKLine ln, CGRect rc);

#pragma mark -
#pragma mark Segments

extern const LKSegment LKSegmentNull;

static inline LKSegment
LKSegmentBetween(CGPoint ptStart, CGPoint ptEnd) 
{
    return (LKSegment){ ptStart, ptEnd };
}

extern BOOL LKSegmentIsNull(LKSegment sg);

static inline LKLine
LKLineFromSegment(LKSegment sg) 
{
    return (LKLine){ sg.ptStart, CGSizeBetween(sg.ptStart, sg.ptEnd) };
}

inline static CGPoint
CGPointOnSegment(LKSegment sg, float t)
{
    return CGPointAdd(sg.ptStart, CGSizeScale(CGSizeBetween(sg.ptStart, sg.ptEnd), t));
}


extern BOOL LKSegmentIntersectsRect(LKSegment sg, CGRect rc);
extern LKSegment LKSegmentIntersectionRect(LKSegment sg, CGRect rc);


#pragma mark -
#pragma mark Angles

inline static float
LKRadFromDeg(float degrees)
{
    return (float)(degrees * M_PI / 180.0);
}

inline static float
LKDegFromRad(float radians)
{
    return (float)(radians * 180.0 / M_PI);
}
