//
//  LKGeometry.m
//  Vectors
//
//  Created by Károly Lőrentey on 2010-10-17.
//  Copyright 2010 Károly Lőrentey. All rights reserved.
//

#import "LKGeometry.h"
#import "LKFloat.h"

#pragma mark -

CGAffineTransform
CGAffineTransformMakeScaleAround(float xScale, float yScale, CGPoint center)
{
#if 0
    CGAffineTransform t = CGAffineTransformIdentity;
    t = CGAffineTransformTranslate(t, center.x, center.y);
    t = CGAffineTransformScale(t, xScale, yScale);
    t = CGAffineTransformTranslate(t, -center.x, -center.y);
#else
    CGAffineTransform t = CGAffineTransformMake(xScale, 0,
                                                0, yScale,
                                                (1 - xScale) * center.x,
                                                (1 - yScale) * center.y);
#endif
    return t;
}

CGFloat
CGPointDistance (CGPoint pt1, CGPoint pt2)
{
    CGFloat dx = pt1.x - pt2.x;
    CGFloat dy = pt1.y - pt2.y;
    return hypotf(dx, dy);
}

CGPoint
CGPointScale(CGPoint pt, float scale)
{
    pt.x *= scale;
    pt.y *= scale;
    return pt;
}

CGPoint
CGPointInterpolate(float t, CGPoint pt1, CGPoint pt2)
{
    return CGPointMake(pt1.x + t * (pt2.x - pt1.x), pt1.y + t * (pt2.y - pt1.y));
}

float
CGSizeLength(CGSize size)
{
    return sqrtf(size.width * size.width + size.height * size.height);
}

CGSize
CGSizeScale(CGSize size, float scale)
{
    size.width *= scale;
    size.height *= scale;
    return size;
}

float
CGSizeAngleBetween(CGSize vk1, CGSize vk2)
{
    float d = vk1.width * vk2.height - vk1.height * vk2.width;
    float angle = acosf(CGSizeDotProduct(vk1, vk2) / (CGSizeLength(vk1) * CGSizeLength(vk2)));
    return (d < 0 ? -angle : angle);
}

CGRect
CGRectRoundToInteger(CGRect rect)
{
    rect.origin.x = roundf(rect.origin.x);
    rect.origin.y = roundf(rect.origin.y);
    rect.size.width = roundf(rect.size.width);
    rect.size.height = roundf(rect.size.height);
    return rect;
}

CGRect
CGRectScaleAroundOrigin(CGRect rect, float scaleX, float scaleY)
{
    rect.origin.x *= scaleX;
    rect.origin.y *= scaleY;
    rect.size.width *= scaleX;
    rect.size.height *= scaleY;
    return rect;
}

CGRect
CGRectScaleAroundMiddle(CGRect rect, float scaleX, float scaleY)
{
    CGPoint middle = CGPointMake(rect.origin.x + rect.size.width / 2,
                                 rect.origin.y + rect.size.height / 2);
    rect.size.width *= scaleX;
    rect.size.height *= scaleY;
    rect.origin.x = middle.x - rect.size.width / 2;
    rect.origin.y = middle.y - rect.size.height / 2;
    return rect;
}

CGRect
CGRectExtendToFitAspect(CGRect rect, CGSize size)
{
    float r = rect.size.width / rect.size.height;
    float ratio = size.width / size.height;
    if (fequalf_abs(r, ratio, 0.001))
        return rect;
    else if (r < ratio) { // Rect is narrower than we want.
        float dx = ratio * rect.size.height - rect.size.width;
        assert(dx > 0);
        return CGRectMake(rect.origin.x - dx / 2, rect.origin.y,
                          ratio * rect.size.height, rect.size.height);
    }
    else { // Rect is wider than we want.
        float dy = rect.size.width / ratio - rect.size.height;
        assert(dy > 0);
        return CGRectMake(rect.origin.x, rect.origin.y - dy / 2,
                          rect.size.width, rect.size.height / ratio);
    }
}

CGRect
CGRectMoveToFitInRect(CGRect rect, CGRect bounds)
{
    // Check right/bottom edges first, so that if rect is wider or taller
    // than bounds, we will position it at the top/left corner.
    if (CGRectGetMaxX(rect) > CGRectGetMaxX(bounds))
        rect.origin.x -= CGRectGetMaxX(rect) - CGRectGetMaxX(bounds);
    if (CGRectGetMaxY(rect) > CGRectGetMaxY(bounds))
        rect.origin.y -= CGRectGetMaxY(rect) - CGRectGetMaxY(bounds);
    if (rect.origin.x < bounds.origin.x)
        rect.origin.x += bounds.origin.x - rect.origin.x;
    if (rect.origin.y < bounds.origin.y)
        rect.origin.y += bounds.origin.y - rect.origin.y;
    return rect;
}

CGPoint
CGRectGetCorner(CGRect rect, CGRectCorner corner)
{
    switch (corner) {
        default:
        case CGRectTopLeft:
            return CGPointMake(rect.origin.x, rect.origin.y);
        case CGRectTopCenter:
            return CGPointMake(rect.origin.x + rect.size.width/2, rect.origin.y);
        case CGRectTopRight:
            return CGPointMake(rect.origin.x + rect.size.width, rect.origin.y);
        case CGRectCenterLeft:
            return CGPointMake(rect.origin.x, rect.origin.y + rect.size.width/2);
        case CGRectCenterCenter:
            return CGPointMake(rect.origin.x + rect.size.width/2, rect.origin.y + rect.size.width/2);
        case CGRectCenterRight:
            return CGPointMake(rect.origin.x + rect.size.width, rect.origin.y + rect.size.width/2);
        case CGRectBottomLeft:
            return CGPointMake(rect.origin.x, rect.origin.y + rect.size.height);
        case CGRectBottomCenter:
            return CGPointMake(rect.origin.x + rect.size.width/2, rect.origin.y + rect.size.height);
        case CGRectBottomRight:
            return CGPointMake(rect.origin.x + rect.size.width, rect.origin.y + rect.size.height);
    }
}

void
CGRectSetCorner(CGRect *rect, CGRectCorner corner, CGPoint point)
{
    switch (corner) {
        default:
        case CGRectTopLeft:
            rect->size.width -= point.x - rect->origin.x;
            rect->origin.x = point.x;
            
            rect->size.height -= point.y - rect->origin.y;
            rect->origin.y = point.y;
            break;
        case CGRectTopCenter:
            rect->size.height -= point.y - rect->origin.y;
            rect->origin.y = point.y;
            break;
        case CGRectTopRight:
            rect->size.width = point.x - rect->origin.x;
            
            rect->size.height -= point.y - rect->origin.y;
            rect->origin.y = point.y;
            break;
        case CGRectCenterLeft:
            rect->size.width -= point.x - rect->origin.x;
            rect->origin.x = point.x;
            break;
        case CGRectCenterCenter:
            rect->origin.x = point.x - rect->size.width / 2;
            rect->origin.y = point.y - rect->size.height / 2;
            break;
        case CGRectCenterRight:
            rect->size.width = point.x - rect->origin.x;
            break;
        case CGRectBottomLeft:
            rect->size.width -= point.x - rect->origin.x;
            rect->origin.x = point.x;
            
            rect->size.height = point.y - rect->origin.y;
            break;
        case CGRectBottomCenter:
            rect->size.height = point.y - rect->origin.y;
            break;
        case CGRectBottomRight:
            rect->size.width = point.x - rect->origin.x;
            rect->size.height = point.y - rect->origin.y;
            break;
    }
}

#pragma mark -
void
CGPathAddRoundRect(CGMutablePathRef path, CGRect rect, float radius)
{
    // Bottom left corner.
	CGPathAddArc(path, NULL, rect.origin.x + radius, rect.origin.y + rect.size.height - radius,
                 radius, (float)M_PI, (float)(M_PI / 2), 1);
    // Bottom right corner.
	CGPathAddArc(path, NULL, rect.origin.x + rect.size.width - radius,
                 rect.origin.y + rect.size.height - radius, radius, (float)(M_PI / 2), 0.0f, 1);
    // Top right corner.
	CGPathAddArc(path, NULL, rect.origin.x + rect.size.width - radius, rect.origin.y + radius,
                 radius, 0.0f, (float)(-M_PI / 2), 1);
    // Top left corner.
	CGPathAddArc(path, NULL, rect.origin.x + radius, rect.origin.y + radius, radius,
                 (float)(-M_PI / 2), (float)M_PI, 1);
    CGPathCloseSubpath(path);
}

void
CGContextAddRoundRect(CGContextRef c, CGRect rect, float radius)
{
    // Bottom left corner.
	CGContextAddArc(c, rect.origin.x + radius, rect.origin.y + rect.size.height - radius,
                    radius, M_PI, M_PI / 2, 1);
    // Bottom right corner.
	CGContextAddArc(c, rect.origin.x + rect.size.width - radius,
                    rect.origin.y + rect.size.height - radius, radius, M_PI / 2, 0.0f, 1);
    // Top right corner.
	CGContextAddArc(c, rect.origin.x + rect.size.width - radius, rect.origin.y + radius,
                    radius, 0.0f, -M_PI / 2, 1);
    // Top left corner.
	CGContextAddArc(c, rect.origin.x + radius, rect.origin.y + radius, radius,
                    -M_PI / 2, M_PI, 1);
    CGContextClosePath(c);
}

BOOL
CGRectInsideRect(CGRect rcInner, CGRect rcOuter)
{
    return (CGRectGetMinX(rcInner) > CGRectGetMinX(rcOuter)
            && CGRectGetMaxX(rcInner) < CGRectGetMaxX(rcOuter)
            && CGRectGetMinY(rcInner) > CGRectGetMinY(rcOuter)
            && CGRectGetMaxY(rcInner) < CGRectGetMaxY(rcOuter));
}

NSString *
CGRectString(CGRect rect)
{
    return [NSString stringWithFormat:@"Rect[%f, %f, %f, %f]", rect.origin.x, rect.origin.y, rect.size.width, rect.size.height];
}

#pragma mark -
#pragma mark Lines

LKLine
LKLineBetween(CGPoint p1, CGPoint p2)
{
    LKLine ln;
    ln.pt = p1;
    ln.vk = CGSizeBetween(p1, p2);
    return ln;
}

// Value of implicit line equation.
static inline float
LKLineValue(LKLine ln, CGPoint pt)
{
    CGSize nm = CGSizeMake(-ln.vk.height, ln.vk.width);
    CGSize vkLP = CGSizeBetween(ln.pt, pt);
    return nm.width * vkLP.width + nm.height * vkLP.height;    
}

BOOL
CGPointIsToTheRight(LKLine ln, CGPoint pt)
{
    return LKLineValue(ln, pt) >= 0;
}


CGPoint
CGPointIntersection(LKLine ln1, LKLine ln2)
{
    double det = ln1.vk.width * ln2.vk.height - ln1.vk.height * ln2.vk.width;
    if (fabs(det) < 0.0001f)
        return CGPointMake(nanf(""), nanf(""));
    double alpha = (ln2.vk.height * (ln2.pt.x - ln1.pt.x) - ln2.vk.width * (ln2.pt.y - ln1.pt.y)) / det;
    CGPoint res = CGPointMake(ln1.pt.x + alpha * ln1.vk.width, ln1.pt.y + alpha * ln1.vk.height);
    return res;
}


CGPoint
CGPointClosestOnLine(LKLine ln, CGPoint pt)
{
    CGSize w = CGSizeBetween(ln.pt, pt);
    CGSize wp = CGSizeScale(w, CGSizeDotProduct(w, ln.vk) / CGSizeDotProduct(ln.vk, ln.vk));
    return CGPointAdd(ln.pt, wp);
}

float
LKLineParameterClosestOnLine(LKLine ln, CGPoint pt)
{
    CGSize w = CGSizeBetween(ln.pt, pt);
    float t = CGSizeDotProduct(w, ln.vk) / CGSizeDotProduct(ln.vk, ln.vk);
    return t;
}

BOOL
LKLineIntersectsRect(LKLine ln, CGRect rc)
{
    float tl = LKLineValue(ln, CGRectGetCorner(rc, CGRectTopLeft));
    float tr = LKLineValue(ln, CGRectGetCorner(rc, CGRectTopRight));
    float bl = LKLineValue(ln, CGRectGetCorner(rc, CGRectBottomLeft));
    float br = LKLineValue(ln, CGRectGetCorner(rc, CGRectBottomRight));
    BOOL right = tl > 0 && tr > 0 && bl > 0 && br > 0;
    BOOL left = tl < 0 && tr < 0 && bl < 0 && br < 0;
    return !left && !right;
}

// Line/rectangle intersection.
//
// Sets tmin and tmax to parameter values for the endpoints of intersection b/w line and rectangle.
// Returns NO if the line does not intersect the rectangle.
static BOOL
LiangBarsky(LKLine ln, CGRect rc, float *tmin, float *tmax)
{
    // Liang-Barsky algorithm
    CGPoint ptTL = CGRectGetCorner(rc, CGRectTopLeft);
    CGPoint ptBR = CGRectGetCorner(rc, CGRectBottomRight);
    
    // Left, right, top, bottom
    float p[4] = { -ln.vk.width, ln.vk.width, -ln.vk.height, ln.vk.height };
    float q[4] = { ln.pt.x - ptTL.x, ptBR.x - ln.pt.x, ln.pt.y - ptTL.y, ptBR.y - ln.pt.y };
    
    *tmin = -INFINITY;
    *tmax = INFINITY;
    for (int i = 0; i < 4; i++) {
        if (p[i] < 0) {
            // Outside --> inside
            *tmin = fmaxf(*tmin, q[i] / p[i]);
        }
        else if (p[i] > 0) {
            // Inside --> outside
            *tmax = fminf(*tmax, q[i] / p[i]);
        }
        else {
            // Segment is parallel to ith side
            if (q[i] < 0)
                return NO;
        }
    }
    if (*tmin > *tmax)
        return NO;
    assert(isfinite(*tmin) && isfinite(*tmax));
    return YES;
}

LKSegment
LKSegmentIntersectionLineRect(LKLine ln, CGRect rc)
{
    float tmin, tmax;
    if (!LiangBarsky(ln, rc, &tmin, &tmax))
        return LKSegmentNull;
    return LKSegmentBetween(CGPointOnLine(ln, tmin), CGPointOnLine(ln, tmax));
}

#pragma mark -
#pragma mark Segments

const LKSegment LKSegmentNull = { { NAN, NAN }, { NAN, NAN } };

BOOL
LKSegmentIsNull(LKSegment sg)
{
    return isnan(sg.ptStart.x);
}

BOOL
LKSegmentIntersectsRect(LKSegment sg, CGRect rc)
{
    if (!LKLineIntersectsRect(LKLineFromSegment(sg), rc))
        return NO;
    
    CGPoint ptTL = CGRectGetCorner(rc, CGRectTopLeft);
    CGPoint ptBR = CGRectGetCorner(rc, CGRectBottomRight);

    if (sg.ptStart.x > ptBR.x && sg.ptEnd.x > ptBR.x)
        return NO;
    if (sg.ptStart.x < ptTL.x && sg.ptEnd.x < ptTL.x)
        return NO;
    if (sg.ptStart.y > ptBR.y && sg.ptEnd.y > ptBR.y)
        return NO;
    if (sg.ptStart.y < ptTL.y && sg.ptEnd.y < ptTL.y)
        return NO;
    return YES;
}
            
LKSegment
LKSegmentIntersectionRect(LKSegment sg, CGRect rc)
{
    float tmin, tmax;
    if (!LiangBarsky(LKLineFromSegment(sg), rc, &tmin, &tmax))
        return LKSegmentNull;
    tmin = fmaxf(0, tmin);
    tmax = fminf(1, tmax);
    if (tmin > tmax)
        return LKSegmentNull;
    return LKSegmentBetween(CGPointOnSegment(sg, tmin), CGPointOnSegment(sg, tmax));
}
