//
//  LKTiledScrollView.m
//  Vectors
//
//  Created by Károly Lőrentey on 2010-03-08.
//  Copyright 2010 Károly Lőrentey. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>

#import "LKTiledCanvasView.h"
#import "LKTileLayer.h"
#import "LKGeometry.h"

//#define BITMAPCONTEXT 1

#define DEFAULT_MIN_ZOOM 0.2
#define DEFAULT_MAX_ZOOM 20

#define _zoomScale (_appliedZoomScale * _fastZoomScale)

// Slightly adjust ptd so that when it's used as the view origin (_rcdVisible.origin),
// all tiles fall on pixel boundaries.
static CGPoint 
CGPointSnapToPixel(CGPoint ptd, float zoomScale)
{
    // We can ignore the current zoom factor and view origin, because
    // we simply need to ensure that point (0,0) in the document will fall on
    // a pixel boundary after ptd is set as the view origin.
    CGPoint ptv = CGPointScale(ptd, zoomScale);
    ptv.x = floorf(ptv.x + 0.5f);
    ptv.y = floorf(ptv.y + 0.5f);
    
    return CGPointScale(ptv, 1 / zoomScale);
}

@interface LKTiledCanvasView ()
- (void)didReceiveMemoryWarning:(NSNotification *)notification;
- (LKTileLayer *)dequeueReusableTile;
- (void)drawTile:(LKTileLayer *)tl rcdClip:(CGRect)rcdClip;
- (void)applyPtdOrigin:(CGPoint)ptdOrigin appliedZoom:(float)appliedZoomScale fastZoom:(float)fastZoomScale;
- (void)setupGestures;
- (void)handleDoubleTapGesture:(UITapGestureRecognizer *)tapper;
- (void)handleDragGesture:(UIPanGestureRecognizer *)dragger;
- (void)handlePinchGesture:(UIPinchGestureRecognizer *)pincher;
@end

@implementation LKTiledCanvasView
@synthesize delegate = _delegate;
@synthesize tileSize = _rszTile;
@synthesize minimumZoomScale = _minZoomScale;
@synthesize maximumZoomScale = _maxZoomScale;
#ifdef DEBUG
@synthesize enableDebugOutput = _enableDebugOutput;
#endif
@synthesize rcdContents = _rcdContents;

- (void)_setupCanvasView {
    self.clipsToBounds = YES;
    self.opaque = NO;
    self.autoresizesSubviews = YES;
    
    _tileCollectorLayer = [[CAScrollLayer alloc] init];
    _tileCollectorLayer.anchorPoint = CGPointZero;
    _tileCollectorLayer.frame = self.bounds;
    _tileCollectorLayer.masksToBounds = NO;
    [self.layer addSublayer:_tileCollectorLayer];
    
    self.tileSize = CGSizeMake(256, 256);
    
    _reusableTiles = [[NSMutableSet alloc] init];
    _tiles = [[NSMutableArray alloc] init];
    
    _firstVisibleRow = _firstVisibleColumn = NSIntegerMax;
    _lastVisibleRow = _lastVisibleColumn = NSIntegerMin;
    
    _rcvLayout = self.bounds;
    _rcdVisible = self.bounds;
    _rcdContents = CGRectNull;
    
    _minZoomScale = DEFAULT_MIN_ZOOM;
    _maxZoomScale = DEFAULT_MAX_ZOOM;
    _appliedZoomScale = 1.0f;
    _fastZoomScale = 1.0f;
    
    NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
    [nc addObserver:self 
           selector:@selector(didReceiveMemoryWarning:) 
               name:UIApplicationDidReceiveMemoryWarningNotification
             object:nil];
    
    [self setupGestures];
}

- (id)initWithCoder:(NSCoder *)aDecoder {
    self = [super initWithCoder:aDecoder];
    if (self == nil)
        return nil;
    [self _setupCanvasView];
    return self;
}


- (id)initWithFrame:(CGRect)frame {
    self = [super initWithFrame:frame];
    if (self == nil)
        return nil;
    [self _setupCanvasView];
    return self;
}

- (void)didReceiveMemoryWarning:(NSNotification *)notification {
    // Clear reusable tiles.
    [_reusableTiles removeAllObjects];
}

- (void)dealloc {
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    LKRelease(_tileCollectorLayer);
    LKRelease(_reusableTiles);
    LKRelease(_tiles);
    
    [super dealloc];
}

#pragma mark -

- (void)setRcdContents:(CGRect)rcdContents {
    if (!CGRectEqualToRect(rcdContents, _rcdContents)) {
        _rcdContents = rcdContents;
        [self setNeedsLayout];
    }
}

- (LKTileLayer *)dequeueReusableTile {
    LKTileLayer *tl = [_reusableTiles anyObject];
    if (tl) {
        [[tl retain] autorelease];
        [_reusableTiles removeObject:tl];
    }
    else {
        tl = [[[LKTileLayer alloc] init] autorelease];
        tl.anchorPoint = CGPointZero;
#ifdef BITMAPCONTEXT
        tl.bitmapContext = [[[LKBitmapContext alloc] initWithWidth:_rszTile.width 
                                                            height:_rszTile.height
                                                             alpha:YES] autorelease];
#endif
        tl.contents = nil;
    }
    return tl;
}

- (void)reloadData {
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    for (LKTileLayer *tl in _tiles) {
        tl.contents = nil;
        [_reusableTiles addObject:tl];
        [tl removeFromSuperlayer];
    }
    [_tiles removeAllObjects];
    assert([_tileCollectorLayer.sublayers count] == 0);
    _firstVisibleRow = _firstVisibleColumn = NSIntegerMax;
    _lastVisibleRow = _lastVisibleColumn = NSIntegerMin;

    [self setNeedsLayout];
    [self layoutIfNeeded];
    [CATransaction commit];
}

#pragma mark -

- (void)layoutSubviews {
    assert(CGPointEqualToPoint(self.bounds.origin, CGPointZero));

    [super layoutSubviews];

    if (!CGRectEqualToRect(_rcvLayout, self.bounds)) {
        // The size of the view has changed; adjust rcdVisible accordingly.

        // The new viewport is centered on the middle of the old.
        CGSize rsdNew = [self rsdFromRsv:self.bounds.size];
        CGPoint ptdCenter = [self ptdCenter];
        CGRect rcdNew = CGRectMake(ptdCenter.x - rsdNew.width / 2,
                                   ptdCenter.y - rsdNew.height / 2,
                                   rsdNew.width, rsdNew.height);

        [self applyPtdOrigin:rcdNew.origin appliedZoom:_appliedZoomScale fastZoom:_fastZoomScale];

        _rcvLayout = self.bounds;
    }

    CGRect rcd = (CGRectIsNull(_rcdContents) ? _rcdVisible : CGRectIntersection(_rcdVisible, _rcdContents));
    if (CGRectIsNull(rcd)) {
        // Whole document is outside our viewport
        rcd = CGRectZero;
    }
    int firstNeededRow = floorf(CGRectGetMinY(rcd) / _rszTile.height * _appliedZoomScale);
    int firstNeededColumn = floorf(CGRectGetMinX(rcd) / _rszTile.width * _appliedZoomScale);
    int lastNeededRow = floorf(CGRectGetMaxY(rcd) / _rszTile.height * _appliedZoomScale);
    int lastNeededColumn = floorf(CGRectGetMaxX(rcd) / _rszTile.width * _appliedZoomScale);

    if (firstNeededRow == _firstVisibleRow
        && firstNeededColumn == _firstVisibleColumn
        && lastNeededRow == _lastVisibleRow
        && lastNeededColumn == _lastVisibleColumn)
        return;
    
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    
    // Remove unnecessary tiles.
    for (int i = [_tiles count] - 1; i >= 0; i--) {
        LKTileLayer *tl = [_tiles objectAtIndex:i];
        int row = tl->_row;
        int col = tl->_column;

        if (row >= firstNeededRow && row <= lastNeededRow
            && col >= firstNeededColumn && col <= lastNeededColumn)
            continue;

        [_reusableTiles addObject:tl];
        [tl removeFromSuperlayer];
        [_tiles removeObjectAtIndex:i];
    }

    // Add missing tiles.
    for (int row = firstNeededRow; row <= lastNeededRow; row++) {
        for (int col = firstNeededColumn; col <= lastNeededColumn; col++) {
            BOOL tileIsMissing = (_firstVisibleRow > row || _firstVisibleColumn > col
                                  || _lastVisibleRow < row || _lastVisibleColumn < col);
            if (tileIsMissing) {
                CGSize rsdTile = CGSizeMake(_rszTile.width / _appliedZoomScale, _rszTile.height / _appliedZoomScale);
                CGPoint ptdTile = CGPointMake(col * rsdTile.width, row * rsdTile.height);
                CGRect rcdTile = CGRectMake(ptdTile.x, ptdTile.y, rsdTile.width, rsdTile.height);

                LKTileLayer *tl = [self dequeueReusableTile];
                tl->_row = row;
                tl->_column = col;

                tl.bounds = rcdTile;
                tl.position = ptdTile;
                [self drawTile:tl rcdClip:CGRectNull];
                
                [_tiles addObject:tl];
                [_tileCollectorLayer addSublayer:tl];
            }
        }
    }
    assert(_tileCollectorLayer.sublayers.count <= (lastNeededRow - firstNeededRow + 1) * (lastNeededColumn - firstNeededColumn + 1));
    [CATransaction commit];
    
    _firstVisibleRow = firstNeededRow;
    _firstVisibleColumn = firstNeededColumn;
    _lastVisibleRow = lastNeededRow;
    _lastVisibleColumn = lastNeededColumn;
}

- (void)drawTile:(LKTileLayer *)tl rcdClip:(CGRect)rcdClip {
    CGRect rcdTile = tl.bounds;
    
#ifdef BITMAPCONTEXT
    CGContextRef c = tl.bitmapContext.CGContext;
#else
    UIGraphicsBeginImageContextWithOptions(_rszTile, NO, 0.0);
    CGContextRef c = UIGraphicsGetCurrentContext();
#endif

    CGContextSaveGState(c);
    CGContextScaleCTM(c, _appliedZoomScale, _appliedZoomScale);
    CGContextTranslateCTM(c, -rcdTile.origin.x, -rcdTile.origin.y);
#ifdef BITMAPCONTEXT
    if (!CGRectIsNull(rcdClip)) {
        CGRect rczClip = CGRectScaleAroundOrigin(rcdClip, _appliedZoomScale, _appliedZoomScale);
        rczClip = CGRectIntegral(rczClip);
        CGContextClipToRect(c, CGRectScaleAroundOrigin(rczClip, 1/_appliedZoomScale, 1/_appliedZoomScale));        
    }
#endif
    [self.delegate drawTileForView:self inContext:c];
    CGContextRestoreGState(c);

#ifdef DEBUG
    if (_enableDebugOutput) {
        // Draw debug overlay
        CGContextSaveGState(c);
        CGContextScaleCTM(c, _appliedZoomScale, _appliedZoomScale);
        CGContextTranslateCTM(c, -rcdTile.origin.x, -rcdTile.origin.y);
        CGContextSetFillColorWithColor(c, [UIColor colorWithHue:(arc4random() % 1000) / 1000.0 saturation:1 brightness:1 alpha:0.15].CGColor);
        if (CGRectIsNull(rcdClip))
            CGContextFillRect(c, rcdTile);
        else
            CGContextFillRect(c, rcdClip);
        CGContextRestoreGState(c);

        // Draw debug label
        CGContextSaveGState(c);
        CGContextSelectFont(c, "Helvetica", 24.0, kCGEncodingMacRoman);
        CGContextSetTextMatrix(c, CGAffineTransformMakeScale(1, -1));
        CGContextSetTextDrawingMode(c, kCGTextFill);
        
        char buf[20];
        sprintf(buf, "#%d (%d,%d)", tl.index, tl.column, tl.row);
        CGContextSetFillColorWithColor(c, [UIColor blackColor].CGColor);
        CGContextShowTextAtPoint(c, 7, 31.0, buf, strlen(buf));
        CGContextSetFillColorWithColor(c, [UIColor greenColor].CGColor);
        CGContextShowTextAtPoint(c, 6, 30.0, buf, strlen(buf));
        
        CGContextRestoreGState(c);
    }
#endif
    

    CGImageRef cgimage = CGBitmapContextCreateImage(c);
    
    tl.contents = (id)cgimage;
    
    CGImageRelease(cgimage);
#ifndef BITMAPCONTEXT
    UIGraphicsEndImageContext();
#endif
}

#pragma mark -
#pragma mark setNeedsDisplay

- (void)setNeedsDisplay {
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    [super setNeedsDisplay];
    for (LKTileLayer *tl in _tiles) {
        [self drawTile:tl rcdClip:CGRectNull];
    }
    [CATransaction commit];
}

- (void)setNeedsDisplayInRcd:(CGRect)rcd {
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    CGRect rcv = [self rcvFromRcd:rcd];
    [super setNeedsDisplayInRect:rcv];
    for (LKTileLayer *tl in _tiles) {
        CGRect rcdTile = tl.bounds;
        if (CGRectIntersectsRect(rcd, rcdTile)) {
            CGRect rcdUpdate = CGRectIntersection(rcd, rcdTile);
            [self drawTile:tl rcdClip:rcdUpdate];
        }
    }
    [CATransaction commit];
}

- (void)setNeedsDisplayInRect:(CGRect)rect {
    [self setNeedsDisplayInRcd:[self rcdFromRcv:rect]];
}

#pragma mark -
#pragma mark Coordinate conversion

- (CGPoint)ptdFromPtv:(CGPoint)ptv {
    return [_tileCollectorLayer convertPoint:ptv fromLayer:self.layer];
}

- (CGPoint)ptvFromPtd:(CGPoint)ptd {
    return [_tileCollectorLayer convertPoint:ptd toLayer:self.layer];
}

- (CGSize)rsdFromRsv:(CGSize)rsv {
    return CGSizeMake(rsv.width / _zoomScale, rsv.height / _zoomScale);
}

- (CGSize)rsvFromRsd:(CGSize)rsd {
    return CGSizeMake(_zoomScale * rsd.width, _zoomScale * rsd.height);
}


- (CGRect)rcdFromRcv:(CGRect)rcv {
    return [_tileCollectorLayer convertRect:rcv fromLayer:self.layer];
}

- (CGRect)rcvFromRcd:(CGRect)rcd {
    return [_tileCollectorLayer convertRect:rcd toLayer:self.layer];
}

- (CGPoint)ptdCenter {
    return CGPointMake(_rcdVisible.origin.x + _rcdVisible.size.width / 2,
                       _rcdVisible.origin.y + _rcdVisible.size.height / 2);
}

- (CGRect)rcdVisible {
    return _rcdVisible;
}

- (float)zoomScale { 
    return _zoomScale;
}

- (CGPoint)ptdOrigin {
    return _rcdVisible.origin;
}


#pragma mark -
#pragma mark Navigation

- (void)applyPtdOrigin:(CGPoint)ptdOrigin appliedZoom:(float)appliedZoomScale fastZoom:(float)fastZoomScale {
    float zoomScale = appliedZoomScale * fastZoomScale;
    
    CGRect rcdVisible = CGRectOffset(CGRectScaleAroundOrigin(self.bounds, 1/zoomScale, 1/zoomScale), 
                                     ptdOrigin.x, ptdOrigin.y);
    
    if (appliedZoomScale != _appliedZoomScale 
        || fastZoomScale != _fastZoomScale
        || !CGRectEqualToRect(rcdVisible, _rcdVisible))
    {
        _appliedZoomScale = appliedZoomScale;
        _fastZoomScale = fastZoomScale;
        _rcdVisible = rcdVisible;
        rcdVisible.origin = CGPointSnapToPixel(rcdVisible.origin, zoomScale);
        _tileCollectorLayer.bounds = _rcdVisible;
        _tileCollectorLayer.transform = CATransform3DMakeScale(zoomScale, zoomScale, 1);
        [self setNeedsLayout];
    }
}

- (void)translateToPoint:(CGPoint)ptd {
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    [self applyPtdOrigin:ptd appliedZoom:_appliedZoomScale fastZoom:_fastZoomScale];
    [CATransaction commit];
}

- (void)translateByVector:(CGSize)vkd {
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    CGPoint ptdOrigin = CGPointSnapToPixel(CGPointAdd(_rcdVisible.origin, vkd), _zoomScale);
    [self applyPtdOrigin:ptdOrigin appliedZoom:_appliedZoomScale fastZoom:_fastZoomScale];
    [CATransaction commit];
}

- (void)commitFastZoom {
    if (_fastZoomScale == 1.0f)
        return;
    
    [self applyPtdOrigin:_rcdVisible.origin appliedZoom:_appliedZoomScale * _fastZoomScale fastZoom:1.0f];
    [self reloadData];
}

- (void)zoomToRcd:(CGRect)rcd fast:(BOOL)fast {
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    // Find rcdNew that best displays rcd without exceeding resolution limits.
    CGSize size = self.bounds.size;
    CGRect rcdNew = CGRectExtendToFitAspect(rcd, size);
    float scale = size.width / rcdNew.size.width;
    
    if (fast) {
        [self applyPtdOrigin:rcdNew.origin appliedZoom:_appliedZoomScale fastZoom:scale / _appliedZoomScale];
        
        [NSObject cancelPreviousPerformRequestsWithTarget:self selector:@selector(commitFastZoom) object:nil];
        [self performSelector:@selector(commitFastZoom) withObject:nil afterDelay:0.1];
    }
    else {
        [self applyPtdOrigin:rcdNew.origin appliedZoom:scale fastZoom:1.0f];
        [self reloadData];
    }
    [CATransaction commit];
}

- (void)zoomToScale:(float)scale aroundPtd:(CGPoint)ptd fast:(BOOL)fast {
    float clampedScale = fclampf(scale, _minZoomScale, _maxZoomScale);
    
    // Don't call zoomToRcd: when we are at the limit of our resolution:
    // it may reposition the view to center it, while we want ptdCenter
    // to remain at a fixed position on screen.
    if (clampedScale == _zoomScale && (fast || _fastZoomScale == 1.0f))
        return;
    
    // Force redraw periodically while zooming out to prevent the need
    // for too many tiles at once.
    if (fast && clampedScale / _appliedZoomScale < 0.66f)
        fast = NO;
    
    CGAffineTransform rcdTransform = CGAffineTransformMakeScaleAround(self.zoomScale / clampedScale,
                                                                      self.zoomScale / clampedScale,
                                                                      ptd);
    CGRect rcdNew = CGRectApplyAffineTransform(self.rcdVisible, rcdTransform);
    [self zoomToRcd:rcdNew fast:fast];
}

#pragma mark - Gesture recognizers

- (void)setupGestures {
    UITapGestureRecognizer *doubleTapper = [[UITapGestureRecognizer alloc]
                                            initWithTarget:self
                                            action:@selector(handleDoubleTapGesture:)];
    doubleTapper.numberOfTapsRequired = 2;
    doubleTapper.delegate = self;
    [self addGestureRecognizer:doubleTapper];
    [doubleTapper release];
    
    UIPinchGestureRecognizer *pincher = [[UIPinchGestureRecognizer alloc]
                                         initWithTarget:self
                                         action:@selector(handlePinchGesture:)];
    pincher.delegate = self;
    [self addGestureRecognizer:pincher];
    [pincher release];
    
    UIPanGestureRecognizer *dragger = [[UIPanGestureRecognizer alloc]
                                       initWithTarget:self
                                       action:@selector(handleDragGesture:)];
    dragger.delegate = self;
    dragger.minimumNumberOfTouches = 1;
    dragger.maximumNumberOfTouches = 2;
    [self addGestureRecognizer:dragger];
    [dragger release];
}

- (BOOL)gestureRecognizer:(UIGestureRecognizer *)gr1 shouldRecognizeSimultaneouslyWithGestureRecognizer:(UIGestureRecognizer *)gr2 {
    // Allow two-finger dragging and pinching to work simultaneously,
    // but make all other recognizers mutually exclusive.
    return ([gr1 isKindOfClass:UIPinchGestureRecognizer.class]
            && [gr2 isKindOfClass:UIPanGestureRecognizer.class]);
}

- (void)handleDoubleTapGesture:(UITapGestureRecognizer *)tapper {
    [self zoomToRcd:self.rcdContents fast:NO];
}

- (void)handleDragGesture:(UIPanGestureRecognizer *)dragger {
    CGPoint ptvDelta = [dragger translationInView:self];
    CGSize vkd = [self rsdFromRsv:CGSizeMake(-ptvDelta.x, -ptvDelta.y)];
    [self translateByVector:vkd];
    [dragger setTranslation:CGPointZero inView:self];
}

- (void)handlePinchGesture:(UIPinchGestureRecognizer *)pincher {
    UIGestureRecognizerState state = pincher.state;
    if (state == UIGestureRecognizerStateBegan) {
    }
    CGPoint ptvCenter = [pincher locationInView:self];
    CGPoint ptdCenter = [self ptdFromPtv:ptvCenter];
    float scale = pincher.scale * [self zoomScale];
    [self zoomToScale:scale aroundPtd:ptdCenter fast:YES];
    pincher.scale = 1.0;
    if (state == UIGestureRecognizerStateEnded || state == UIGestureRecognizerStateCancelled) {
        [self commitFastZoom];
    }
}



@end
